

<hr />
<p>
As a sender, you may opt out of having receving notifications
from this site.  If you do, you won't know when your messages
get quarantined and you won't be able to release your messages
from quarantine.
</p><p>
Depending on the situation, the recipient may get notification
that a message was held for them.  If so there is some chance that
the recipient may release the message from quarantine.
</p><p>
<a href="[% GET sender_url %]">Click here</a> to change your sender options.
</p>
